@@ -1,4 +1,4 @@
-# Copyright 2008-2013 Canonical Ltd.  All rights reserved.
+# Copyright 2008-2018 Canonical Ltd.  All rights reserved.
 
 # This file is part of wadllib.
 #
@@ -41,38 +41,20 @@ __all__ = [
     'WADLError',
     ]
 
-try:
-    # Make sure to try the Python 2 form first; we don't want to use
-    # io.StringIO in Python 2, because it is strict about only accepting
-    # unicode input.
-    from cStringIO import StringIO as BytesIO
-except ImportError:
-    from io import BytesIO
 import datetime
-try:
-    from email.mime.multipart import MIMEMultipart
-    from email.mime.nonmultipart import MIMENonMultipart
-except ImportError:
-    # We must be on 2.4.
-    from email.MIMEMultipart import MIMEMultipart
-    from email.MIMENonMultipart import MIMENonMultipart
-
+from email.utils import quote
+import io
+import json
+import random
+import re
+import sys
 import time
 try:
     from urllib.parse import urlencode
 except ImportError:
     from urllib import urlencode
-try:
-    import simplejson as json
-except ImportError:
-    import json
-try:
-    import xml.etree.cElementTree as ET
-except ImportError:
-    try:
-        import cElementTree as ET
-    except ImportError:
-        import elementtree.ElementTree as ET
+import xml.etree.cElementTree as ET
+
 from lazr.uri import URI, merge
 
 from wadllib import (
@@ -823,6 +805,105 @@ class RepresentationDefinition(WADLResolvableDefinition, HasParametersMixin):
         """The media type of the representation described here."""
         return self.resolve_definition().tag.attrib['mediaType']
 
+    def _make_boundary(self, all_parts):
+        """Make a random boundary that does not appear in `all_parts`."""
+        _width = len(repr(sys.maxsize - 1))
+        _fmt = '%%0%dd' % _width
+        token = random.randrange(sys.maxsize)
+        boundary = ('=' * 15) + (_fmt % token) + '=='
+        if all_parts is None:
+            return boundary
+        b = boundary
+        counter = 0
+        while True:
+            pattern = ('^--' + re.escape(b) + '(--)?$').encode('ascii')
+            if not re.search(pattern, all_parts, flags=re.MULTILINE):
+                break
+            b = boundary + '.' + str(counter)
+            counter += 1
+        return b
+
+    def _write_headers(self, buf, headers):
+        """Write MIME headers to a file object."""
+        for key, value in headers:
+            buf.write(key.encode('UTF-8'))
+            buf.write(b': ')
+            buf.write(value.encode('UTF-8'))
+            buf.write(b'\r\n')
+        buf.write(b'\r\n')
+
+    def _write_boundary(self, buf, boundary, closing=False):
+        """Write a multipart boundary to a file object."""
+        buf.write(b'--')
+        buf.write(boundary.encode('UTF-8'))
+        if closing:
+            buf.write(b'--')
+        buf.write(b'\r\n')
+
+    def _generate_multipart_form(self, parts):
+        """Generate a multipart/form-data message.
+
+        This is very loosely based on the email module in the Python standard
+        library.  However, that module doesn't really support directly embedding
+        binary data in a form: various versions of Python have mangled line
+        separators in different ways, and none of them get it quite right.
+        Since we only need a tiny subset of MIME here, it's easier to implement
+        it ourselves.
+
+        :return: a tuple of two elements: the Content-Type of the message, and
+            the entire encoded message as a byte string.
+        """
+        # Generate the subparts first so that we can calculate a safe boundary.
+        encoded_parts = []
+        for is_binary, name, value in parts:
+            buf = io.BytesIO()
+            if is_binary:
+                ctype = 'application/octet-stream'
+                # RFC 7578 says that the filename parameter isn't mandatory
+                # in our case, but without it cgi.FieldStorage tries to
+                # decode as text on Python 3.
+                cdisp = 'form-data; name="%s"; filename="%s"' % (
+                    quote(name), quote(name))
+            else:
+                ctype = 'text/plain; charset="utf-8"'
+                cdisp = 'form-data; name="%s"' % quote(name)
+            self._write_headers(buf, [
+                ('MIME-Version', '1.0'),
+                ('Content-Type', ctype),
+                ('Content-Disposition', cdisp),
+                ])
+            if is_binary:
+                if not isinstance(value, bytes):
+                    raise TypeError('bytes payload expected: %s' % type(value))
+                buf.write(value)
+            else:
+                if not isinstance(value, str):
+                    raise TypeError('str payload expected: %s' % type(value))
+                lines = re.split(r'\r\n|\r|\n', value)
+                for line in lines[:-1]:
+                    buf.write(line.encode('UTF-8'))
+                    buf.write(b'\r\n')
+                buf.write(lines[-1].encode('UTF-8'))
+            encoded_parts.append(buf.getvalue())
+
+        # Create a suitable boundary.
+        boundary = self._make_boundary(b'\r\n'.join(encoded_parts))
+
+        # Now we can write the multipart headers, followed by all the parts.
+        buf = io.BytesIO()
+        ctype = 'multipart/form-data; boundary="%s"' % quote(boundary)
+        self._write_headers(buf, [
+            ('MIME-Version', '1.0'),
+            ('Content-Type', ctype),
+            ])
+        for encoded_part in encoded_parts:
+            self._write_boundary(buf, boundary)
+            buf.write(encoded_part)
+            buf.write(b'\r\n')
+        self._write_boundary(buf, boundary, closing=True)
+
+        return ctype, buf.getvalue()
+
     def bind(self, param_values, **kw_param_values):
         """Bind the definition to parameter values, creating a document.
 
@@ -836,30 +917,13 @@ class RepresentationDefinition(WADLResolvableDefinition, HasParametersMixin):
         if media_type == 'application/x-www-form-urlencoded':
             doc = urlencode(sorted(validated_values.items()))
         elif media_type == 'multipart/form-data':
-            outer = MIMEMultipart()
-            outer.set_type('multipart/form-data')
+            parts = []
             missing = object()
             for param in params:
                 value = validated_values.get(param.name, missing)
                 if value is not missing:
-                    if param.type == 'binary':
-                        maintype, subtype = 'application', 'octet-stream'
-                        params = {}
-                    else:
-                        maintype, subtype = 'text', 'plain'
-                        params = {'charset' : 'utf-8'}
-                    inner = MIMENonMultipart(maintype, subtype, **params)
-                    inner.set_payload(value)
-                    inner['Content-Disposition'] = (
-                        'form-data; name="%s"' % param.name)
-                    outer.attach(inner)
-            doc = str(outer)
-            # Chop off the 'From' line, which only makes sense in an
-            # email.  (Python 3 does not include it.)
-            if doc.startswith('From '):
-                doc = doc[doc.find('\n')+1:]
-            media_type = (outer.get_content_type() +
-                          '; boundary="%s"' % outer.get_boundary())
+                    parts.append((param.type == 'binary', param.name, value))
+            media_type, doc = self._generate_multipart_form(parts)
         elif media_type == 'application/json':
             doc = json.dumps(validated_values)
         else:
@@ -1105,7 +1169,9 @@ class Application(WADLBase):
 
     def _from_string(self, markup):
         """Turns markup into a document."""
-        return self._from_stream(BytesIO(markup))
+        if not isinstance(markup, bytes):
+            markup = markup.encode("UTF-8")
+        return self._from_stream(io.BytesIO(markup))
 
     def get_resource_type(self, resource_type_url):
         """Retrieve a resource type by the URL of its description."""
